/*
 * application.cpp
 *
 * This code handles the application's GUI.
 *
 * This file is distributed as part of Kroket.
 * Copyright (c) 2010 Nicolas BENOIT
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <QtGui>
#include "application.h"
#include "help.h"


/*
 * Constructor
 */
AppKroket::AppKroket ( QWidget *parent )
    : QMainWindow(parent)
{
    ui.setupUi ( this );

    ui.graphicsView->setApp ( this );

    QObject::connect ( ui.actionOpen, SIGNAL(triggered()), this, SLOT(open()) );
    QObject::connect ( ui.actionExportImage, SIGNAL(triggered()), this, SLOT(exportImage()) );
    QObject::connect ( ui.actionPrint, SIGNAL(triggered()), this, SLOT(print()) );

    QObject::connect ( ui.actionUndo, SIGNAL(triggered()), ui.graphicsView, SLOT(undo()) );
    QObject::connect ( ui.actionRedo, SIGNAL(triggered()), ui.graphicsView, SLOT(redo()) );

    QObject::connect ( ui.actionZoomIn, SIGNAL(triggered()), ui.graphicsView, SLOT(zoomIn()) );
    QObject::connect ( ui.actionZoomOut, SIGNAL(triggered()), ui.graphicsView, SLOT(zoomOut()) );
    QObject::connect ( ui.actionZoomFit, SIGNAL(triggered()), ui.graphicsView, SLOT(zoomFit()) );

    QObject::connect ( ui.actionExpandAll, SIGNAL(triggered()), ui.graphicsView, SLOT(expandAll()) );
    QObject::connect ( ui.actionCollapseAll, SIGNAL(triggered()), ui.graphicsView, SLOT(collapseAll()) );

    QObject::connect ( ui.actionReset, SIGNAL(triggered()), ui.graphicsView, SLOT(reset()) );

    QObject::connect ( ui.actionFocus, SIGNAL(triggered()), this, SLOT(focus()) );
    QObject::connect ( ui.actionTrace, SIGNAL(triggered()), this, SLOT(trace()) );
    QObject::connect ( ui.actionFilterTags, SIGNAL(triggered()), this, SLOT(filter()) );

    QObject::connect ( ui.actionHelp, SIGNAL(triggered()), this, SLOT(help()) );
    QObject::connect ( ui.actionHelpInput, SIGNAL(triggered()), this, SLOT(help_input()) );
    QObject::connect ( ui.actionHelpAbout, SIGNAL(triggered()), this, SLOT(help_about()) );

    this->hd = new HelpDialog ( this );
    this->hid = new HelpInputDialog ( this );
    this->had = new HelpAboutDialog ( this );

    this->updateActions ( );
}


/*
 * Callback for the Open signal.
 */
void
AppKroket::open ( )
{
  QString fname = QFileDialog::getOpenFileName ( this, "Open Kroket File", "", "Krokets (*.kk)" );

  if ( fname.size() > 0 )
    {
      ui.graphicsView->loadGraphFromFile ( fname.toAscii() );
      ui.graphicsView->synchronizeView ( );
      this->setWindowTitle ( "kroket - " + ui.graphicsView->getTitle() );
    }
}


/*
 * Callback for the Export Image signal.
 */
void
AppKroket::exportImage ( )
{
  QString fname = QFileDialog::getSaveFileName ( this, "Save Image", "", "Images (*.png)" );

  if ( fname.size() > 0 )
    {
      QImage *img;

      img = ui.graphicsView->getImage ( );
      img->save ( fname, "PNG" );
      delete img;

      QString msg;
      msg.sprintf ( "\'%s\' saved.", qPrintable(fname) );
      ui.statusbar->showMessage ( msg, 2000 );
    }
}


/*
 * Callback for the Print signal.
 */
void
AppKroket::print ( )
{
  QPrinter *printer = new QPrinter ( );
  QPrintDialog *pd = new QPrintDialog ( printer, this );

  if ( pd->exec() == QDialog::Accepted )
    {
      ui.graphicsView->print ( printer );

      QString msg ( "Document printed." );
      ui.statusbar->showMessage ( msg, 2000 );
    }

  delete pd;
  delete printer;
}


/*
 * Callback for the Focus signal.
 */
void
AppKroket::focus ( )
{
  QString nname = ui.txtEntry->text ( );

  if ( nname.size() > 0 )
    {
      if ( !ui.graphicsView->focusOn ( nname ) )
        {
          QString msg;
          msg.sprintf ( "error: \'%s\' could not be found.", qPrintable(nname) );
          ui.statusbar->showMessage ( msg, 2000 );
        }
    }
}


/*
 * Callback for the Trace signal.
 */
void
AppKroket::trace ( )
{
  QString nname = ui.txtEntry->text ( );

  if ( nname.size() > 0 )
    {
      if ( !ui.graphicsView->traceFrom ( nname ) )
        {
          QString msg;
          msg.sprintf ( "error: \'%s\' could not be found.", qPrintable(nname) );
          ui.statusbar->showMessage ( msg, 2000 );
        }
    }
}


/*
 * Callback for the Filter signal.
 */
void
AppKroket::filter ( )
{
  QString tag = ui.txtEntry->text ( );

  if ( tag.size() > 0 )
    {
      if ( tag[0] == '!' )
        {
          tag = tag.mid ( 1 );
          ui.graphicsView->filterWithoutTag ( tag );
        }
      else
        ui.graphicsView->filterWithTag ( tag );
    }
}


/*
 * Callback for the Help signal.
 */
void
AppKroket::help ( )
{
  this->hd->show ( );
}


/*
 * Callback for the HelpInput signal.
 */
void
AppKroket::help_input ( )
{
  this->hid->show ( );
}


/*
 * Callback for the HelpAbout signal.
 */
void
AppKroket::help_about ( )
{
  this->had->show ( );
}


/*
 * Update undo actions buttons and menu entries.
 */
void
AppKroket::updateUndoActions ( )
{
  if ( ui.graphicsView->getUndoStack()->canUndo() )
    ui.actionUndo->setEnabled ( true );
  else
    ui.actionUndo->setEnabled ( false );

  if ( ui.graphicsView->getUndoStack()->canRedo() )
    ui.actionRedo->setEnabled ( true );
  else
    ui.actionRedo->setEnabled ( false );
}


/*
 * Update actions buttons and menu entries.
 */
void
AppKroket::updateActions ( )
{
  this->updateUndoActions ( );

  if ( ui.graphicsView->getGraph() == NULL )
    {
      ui.actionExportImage->setEnabled ( false );
      ui.actionPrint->setEnabled ( false );
      ui.actionZoomIn->setEnabled ( false );
      ui.actionZoomOut->setEnabled ( false );
      ui.actionZoomFit->setEnabled ( false );
      ui.actionExpandAll->setEnabled ( false );
      ui.actionCollapseAll->setEnabled ( false );
      ui.actionReset->setEnabled ( false );
      ui.txtEntry->setEnabled ( false );
      ui.btnFocus->setEnabled ( false );
      ui.btnTrace->setEnabled ( false );
      ui.btnFilter->setEnabled ( false );
    }
  else
    {
      ui.actionExportImage->setEnabled ( true );
      ui.actionPrint->setEnabled ( true );
      ui.actionZoomIn->setEnabled ( true );
      ui.actionZoomOut->setEnabled ( true );
      ui.actionZoomFit->setEnabled ( true );
      ui.actionExpandAll->setEnabled ( true );
      ui.actionCollapseAll->setEnabled ( true );
      ui.actionReset->setEnabled ( true );
      ui.txtEntry->setEnabled ( true );
      ui.btnFocus->setEnabled ( true );
      ui.btnTrace->setEnabled ( true );
      ui.btnFilter->setEnabled ( true );
    }
}


/*
 * Destructor
 */
AppKroket::~AppKroket ( )
{
  delete this->hd;

  /* unregister the app to avoid an unfortunate call to updateAction() */
  ui.graphicsView->setApp ( NULL );
}
